//==========================================================================
//
//      if_s3c4510.c
//
//      Samsung S3C4510x Ethernet Controller device driver.
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    gthomas
// Contributors: gthomas, jskov
//               Grant Edwards <grante@visi.com>
// Date:         2001-07-31
// Purpose:
// Description:
//
//####DESCRIPTIONEND####
//
//========================================================================*/

#include <pkgconf/system.h>
#include <pkgconf/io_eth_drivers.h>
#include CYGHWR_DEVS_ETH_ARM_S3C4510_CFG      // platform specific configuration

#if defined(CYGPKG_REDBOOT)
# include <pkgconf/redboot.h>
#endif

#include <cyg/io/eth/eth_drv.h>
#include <cyg/io/eth/netdev.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>
#include <errno.h>

#include "s3c4510_eth.h"

// Set up the level of debug output
#if CYGDBG_DEVS_ETH_ARM_S3C4510_DEBUG_LEVEL > 0
# define debug1_printf(args...) diag_printf(args)
#else
# define debug1_printf(args...) CYG_EMPTY_STATEMENT
#endif
#if CYGDBG_DEVS_ETH_ARM_S3C4510_DEBUG_LEVEL > 1
# define debug2_printf(args...) diag_printf(args)
#else
# define debug2_printf(args...) CYG_EMPTY_STATEMENT
#endif


//======================================================================
//  Driver statistics

#ifdef CYGDBG_DEVS_ETH_ARM_S3C4510_DRIVER_STATS
//----------------------------------------------------------------------
// Non-SNMP driver statistics

typedef struct s3c4510_drv_stats_s {
    // interrupt entry counters
    cyg_uint32 MAC_Tx_Cnt;
    cyg_uint32 MAC_Rx_Cnt;
    cyg_uint32 BDMA_Tx_Cnt;
    cyg_uint32 BDMA_Rx_Cnt;
    cyg_uint32 Deliver_Cnt;
    cyg_uint32 Poll_Cnt;
    cyg_uint32 tx_send_cnt;
    cyg_uint32 tx_buf_cnt;
    cyg_uint32 rx_buf_cnt;

    // Tx frame descriptor
    cyg_uint32 TxAllocFail;             // Tx buffer alloc failure

    // BDMA Tx
    cyg_uint32 BTxCCP;                  // control frame sent
    cyg_uint32 BTxNLErr;                // BDMATXPTR is null error
    cyg_uint32 BTxNOErr;                // Tx BDMA not owner error
    cyg_uint32 BTxEmptyErr;             // Tx buffer empty error

    // Tx Status
    cyg_uint32 TxPaused;                // Tx paused
    cyg_uint32 TxDefer;                 // Tx deferal
    cyg_uint32 TxParErr;                // Tx parity error
    cyg_uint32 TxHalted;                // Tx halted
    
    // BDMA Rx
    cyg_uint32 BRxRDF;                  // Rx frame moved to memory
    cyg_uint32 BRxNLErr;                // BDMARXPTR is null error
    cyg_uint32 BRxNOErr;                // Rx BDMA not owner error
    cyg_uint32 BRxMSOErr;               // Rx frame too big error
    cyg_uint32 BRxEmptyErr;             // Rx buffer empty error
    cyg_uint32 BRxSEarly;               // Rx early notification
    cyg_uint32 BRxFRF;                  // additional frame received
    cyg_uint32 BRxNRF[0x100];           // # frames in Rx buffer

    // Rx Status
    cyg_uint32 RxHalted;                // Rx halted
    cyg_uint32 RxCtl;                   // Rx control frame
} s3c4510_drv_stats_t;

static s3c4510_drv_stats_t s3c4510_drv_stats = { 0, };


# define STAT_DRV_TST_INC(_tst, _s, _c) \
    if (_tst) STAT_DRV_INC(_s, _c)

# define STAT_DRV_INC(_s, _c)   (s3c4510_drv_stats._s += _c)
# define STAT_DRV_SET(_s, _c)   (s3c4510_drv_stats._s = _c)

#else
# define STAT_DRV_TST_INC(_tst, _s, _c) CYG_EMPTY_STATEMENT
# define STAT_DRV_INC(_s, _c)           CYG_EMPTY_STATEMENT
# define STAT_DRV_SET(_s, _c)           CYG_EMPTY_STATEMENT
#endif // Non-SNMP driver statistics


#if defined(ETH_DRV_GET_IF_STATS) || defined (ETH_DRV_GET_IF_STATS_UD)
//----------------------------------------------------------------------
// SNMP driver statistics

static struct ether_drv_stats s3c4510_snmp_stats = {
    .description = { "eCos S3C4510x Ethernet Controller" },
    .snmp_chipset = { "S3C4510x" },
    .supports_dot3 = true,
};


# define STAT_SNMP_TST_INC(_tst, _s, _c) \
    if (_tst) STAT_SNMP_INC(_s, _c)

# define STAT_SNMP_INC(_s, _c)  (s3c4510_snmp_stats._s += _c)
# define STAT_SNMP_SET(_s, _c)  (s3c4510_snmp_stats._s = _c)

#else
# define STAT_SNMP_TST_INC(_tst, _s, _c) CYG_EMPTY_STATEMENT
# define STAT_SNMP_INC(_s, _c)           CYG_EMPTY_STATEMENT
# define STAT_SNMP_SET(_s, _c)           CYG_EMPTY_STATEMENT
#endif


//======================================================================
//  Include Station Management PHY support

#ifdef CYGPKG_DEVS_ETH_ARM_S3C4510_PHY_NONE
# define PhyReset()     CYG_EMPTY_STATEMENT
#else
# include "phy.inl"
# include CYGHWR_DEVS_ETH_ARM_S3C4510_PHY_INL
#endif


//======================================================================
// Include the platform specific support

#include CYGHWR_DEVS_ETH_ARM_S3C4510_INL

//----------------------------------------------------------------------
//  Get the Ethernet Station (MAC) Address.

static cyg_bool
s3c4510_eth_get_esa(struct eth_drv_sc *sc)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;
    cyg_bool esa_configured = false;

    debug1_printf("s3c4510_eth_get_esa()\n");
    CYG_ASSERTC(spd->pSelf == spd);

#ifdef CYGFUN_ESA_PLF_GET
    esa_configured = CYGFUN_ESA_PLF_GET(spd);
    if (esa_configured) {
        debug1_printf("Got ESA from platform support routine\n");
    }
#endif

#ifdef CYGFUN_ESA_REDBOOT_GET
    if (!esa_configured) {
        esa_configured = CYGFUN_ESA_REDBOOT_GET(spd);
        if (esa_configured) {
            debug1_printf("Got ESA from Redboot option\n");
        }
    }
#endif

    if (!esa_configured && spd->hardwired_esa) {
        // ESA is already set in spd->esa[]
        esa_configured = true;
    }

    if (esa_configured) {
        debug1_printf("ESA: %02x:%02x:%02x:%02x:%02x:%02x\n",
            spd->esa[0], spd->esa[1], spd->esa[2],
            spd->esa[3], spd->esa[4], spd->esa[5]);
    } else {
        diag_printf("ESA was not configured\n");
    }

    return esa_configured;
}


//======================================================================
//  Ethernet buffer data structures and functions

#if defined(CYGPKG_REDBOOT)
# define MAX_RX_BUFFERS         CYGNUM_IO_ETH_DRIVERS_NUM_PKT
# define MAX_TX_BUFFERS         CYGNUM_IO_ETH_DRIVERS_NUM_PKT
#else
# define MAX_RX_BUFFERS         32
# define MAX_TX_BUFFERS         32
#endif


// Ethernet buffers
static s3c4510_eth_buffer_t s3c4510_rx_bufs[MAX_RX_BUFFERS];
static s3c4510_eth_buffer_t s3c4510_tx_bufs[MAX_TX_BUFFERS];


#ifndef CYGARC_UNCACHED_ADDRESS
# define CYGARC_UNCACHED_ADDRESS(x)     ((void *)((cyg_uint32)(x) | (0x1 << 26)))
# define CYGARC_CACHED_ADDRESS(x)       ((void *)((cyg_uint32)(x) & ~(0x1 << 26)))
#endif

//----------------------------------------------------------------------
//  Initialize the Ethernet buffers by linking the frame descriptors
//  into circular lists and pointing the frame descriptors to the
//  Ethernet data buffers.

static void
s3c4510_eth_buf_init(struct eth_drv_sc *sc)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;
    s3c4510_eth_buffer_t *pBuf;
    size_t i;

    debug2_printf("s3c4510_eth_buf_init()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    // link the Tx buffers
    pBuf = (s3c4510_eth_buffer_t *)CYGARC_UNCACHED_ADDRESS(&s3c4510_tx_bufs[0]);
    for (i = 0; i < MAX_TX_BUFFERS; ++i) {
        pBuf[i].fd.FrameDataPtr = (cyg_uint32)&pBuf[0].data[0];
        pBuf[i].fd.Reserved = 0;
        pBuf[i].fd.StatusAndFrameLength = 0;
        pBuf[i].fd.NextFD = (cyg_uint32)&pBuf[i + 1].fd;

        // CYG_ASSERT tag
        pBuf[i].pSelf = &pBuf[i];
    }
    pBuf[MAX_TX_BUFFERS - 1].fd.NextFD = (cyg_uint32)&pBuf[0].fd;

    spd->pTxDone = spd->pTxWrite = &pBuf[0];
    STAT_SNMP_SET(tx_queue_len, MAX_TX_BUFFERS);

    // link the Rx buffers
    pBuf = (s3c4510_eth_buffer_t *)CYGARC_UNCACHED_ADDRESS(&s3c4510_rx_bufs[0]);
    for (i = 0; i < MAX_RX_BUFFERS; ++i) {
        pBuf[i].fd.FrameDataPtr = (cyg_uint32)&pBuf[i].data[0] | FRM_OWNERSHIP_BDMA;
        pBuf[i].fd.Reserved = 0;
        pBuf[i].fd.StatusAndFrameLength = 0;
        pBuf[i].fd.NextFD = (cyg_uint32)&pBuf[i + 1].fd;

        // CYG_ASSERT tag
        pBuf[i].pSelf = &pBuf[i];
    }
    pBuf[MAX_RX_BUFFERS - 1].fd.NextFD = (cyg_uint32)&pBuf[0].fd;

    spd->pRxRead = &pBuf[0];
}


//----------------------------------------------------------------------
//  Record Tx statistics and inform the high-level driver that the
//  packet was transmitted.

static void
s3c4510_eth_process_tx_buf(struct eth_drv_sc *sc, s3c4510_eth_buffer_t *pTxBuf)
{
    cyg_uint32 txStatus = pTxBuf->fd.StatusAndFrameLength >> 16;
    cyg_uint32 collisionCnt = txStatus & 0x1f;
    int status;

    debug2_printf("s3c4510_eth_process_tx_buf()\n");
    STAT_DRV_INC(tx_buf_cnt, 1);
    STAT_SNMP_INC(tx_complete, 1);

    if (txStatus & TXFDSTAT_ERR) {
        // transmit failed, log errors
        STAT_SNMP_TST_INC((txStatus & TXFDSTAT_EX_COLL), tx_max_collisions, 1);
        STAT_SNMP_TST_INC((txStatus & TXFDSTAT_DEFERRED), tx_deferred, 1);
        STAT_SNMP_TST_INC((txStatus & TXFDSTAT_UNDER), tx_underrun, 1);
        STAT_SNMP_TST_INC((txStatus & TXFDSTAT_NCARR), tx_carrier_loss, 1);
        STAT_SNMP_TST_INC((txStatus & TXFDSTAT_SQ_ERR), tx_sqetesterrors, 1);
        STAT_SNMP_TST_INC((txStatus & TXFDSTAT_LATE_COLL), tx_late_collisions, 1);

        STAT_DRV_TST_INC((txStatus & TXFDSTAT_PAUSED), TxPaused, 1);
        STAT_DRV_TST_INC((txStatus & TXFDSTAT_DEFER), TxDefer, 1);
        STAT_DRV_TST_INC((txStatus & TXFDSTAT_PAR), TxParErr, 1);
        STAT_DRV_TST_INC((txStatus & TXFDSTAT_HALTED), TxHalted, 1);

        STAT_SNMP_INC(tx_dropped, 1);
        status = 1;
    } else {
        // transmit OK
        STAT_SNMP_INC(tx_good, 1);
        status = 0;
    }

    if (collisionCnt) {
        if (collisionCnt == 1)
            STAT_SNMP_INC(tx_single_collisions, 1);
        else
            STAT_SNMP_INC(tx_mult_collisions, 1);

        STAT_SNMP_INC(tx_total_collisions, collisionCnt);
    }

    // tell the high-level driver this packet has been transmitted
    sc->funs->eth_drv->tx_done(sc, pTxBuf->key, status);
}


//----------------------------------------------------------------------
//  Record Rx statistics and inform the high-level driver that a
//  packet was received.

static void
s3c4510_eth_process_rx_buf(struct eth_drv_sc *sc, s3c4510_eth_buffer_t *pRxBuf)
{
    cyg_uint32 rxStatus = pRxBuf->fd.StatusAndFrameLength >> 16;

    debug2_printf("s3c4510_eth_process_rx_buf()\n");
    STAT_DRV_INC(rx_buf_cnt, 1);
    STAT_SNMP_INC(rx_count, 1);

    if (rxStatus & RXFDSTAT_ERR) {
        STAT_SNMP_TST_INC((rxStatus & RXFDSTAT_CRC_ERR), rx_crc_errors, 1);
        STAT_SNMP_TST_INC((rxStatus & RXFDSTAT_OV_MAX), rx_resource, 1);
        STAT_SNMP_TST_INC((rxStatus & RXFDSTAT_ALIGN_ERR), rx_align_errors, 1);
        STAT_SNMP_TST_INC((rxStatus & RXFDSTAT_OVERFLOW), rx_overrun_errors, 1);
        STAT_SNMP_TST_INC((rxStatus & RXFDSTAT_LONG_ERR), rx_too_long_frames, 1);
        STAT_SNMP_TST_INC((rxStatus & RXFDSTAT_PAR), rx_symbol_errors, 1);
        STAT_DRV_TST_INC((rxStatus & RXFDSTAT_HALTED), RxHalted, 1);
    }
    else if (rxStatus & RXFDSTAT_GOOD) {
        // receive OK
        cyg_uint32 len = pRxBuf->fd.StatusAndFrameLength & 0xffff;

        STAT_SNMP_INC(rx_good, 1);
        STAT_DRV_TST_INC((rxStatus & RXFDSTAT_CTL_RECD), RxCtl, 1);

        // inform the high-level driver that a packet has arrived, which then
        // will call s3c4510_eth_recv()
        sc->funs->eth_drv->recv(sc, len);
    }
    else {
        diag_printf
            ("s3c4510_process_rx_buf: Rx non-good, non-error frame (%x)\n",
            rxStatus);
    }
}


//----------------------------------------------------------------------
//  Load a MAC address into the CAM and enable it.

static void
s3c4510_eth_set_cam(unsigned index, unsigned char * pAddr)
{
    volatile cyg_uint32 *pCamEntry = &CAM_BASE + index*6/4;

    debug2_printf("s3c4510_eth_set_cam()\n");

    if (index%2 == 0) {
        pCamEntry[0] = (pAddr[0] << 24) | (pAddr[1] << 16) | (pAddr[2] << 8) | pAddr[3];
        pCamEntry[1] = (pCamEntry[1] & 0x0000ffff) | (pAddr[4] << 24) | (pAddr[5] << 16);
    } else {
        pCamEntry[0] = (pCamEntry[0] & 0xffff0000) | (pAddr[0] << 8) | pAddr[1];
        pCamEntry[1] = (pAddr[2] << 24) | (pAddr[3] << 16) | (pAddr[4] << 8) | pAddr[5];
    }

    // enable the CAM entry
    CAMEN |= (0x1 << index);
}


//----------------------------------------------------------------------
//  Initialize the Ethernet controller.

static void
s3c4510_eth_ctrl_init(struct eth_drv_sc *sc)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;

    debug1_printf("s3c4510_eth_ctrl_init()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    PhyReset();
    spd->active = false;
    STAT_SNMP_SET(operational, 2);          // 1 = UNKNOWN, 2 = DOWN, 3 = UP

    // initialize the buffer DMA layer
    BDMARXCON = BDMARXCON_RESET;
    BDMATXCON = BDMATXCON_RESET;
    BDMARXLSZ = MAX_ETH_FRAME_SIZE;

    BDMARXPTR = (cyg_uint32)CYGARC_UNCACHED_ADDRESS(&spd->pRxRead->fd);
    BDMATXPTR = (cyg_uint32)CYGARC_UNCACHED_ADDRESS(&spd->pTxWrite->fd);

    // initialize the MAC layer
    MACCON = MACON_SW_RESET;
    MACCON = MACON_FULL_DUP;
    STAT_SNMP_SET(duplex, 3);               // 1 = UNKNOWN, 2 = SIMPLEX, 3 = DUPLEX
    STAT_SNMP_SET(speed, 100*1000*1000);

    // set up our MAC address
    s3c4510_eth_set_cam(0, spd->esa);

    // enable MAC address filtering and accept broadcast packets
    CAMCON = CAMCON_COMP_EN | CAMCON_BROAD_ACC;

    // configure Rx BDMA
    BDMARXCON =
        BDMARXCON_RX_EMPTY             // Buffer Empty interrupt enable
        | BDMARXCON_EN                 // Enable
        | BDMARXCON_WA00               // Word Alignment- no invalid byte
#if (CYG_BYTEORDER == CYG_MSBFIRST)
        | BDMARXCON_BIG                // Big-endian mode
#else
        | BDMARXCON_LITTLE             // Little-endian mode
#endif
        | BDMARXCON_MSOIE              // Maximum Size Over interrupt enable
        | BDMARXCON_NOIE               // Not Owner interrupt enable
        | BDMARXCON_NLIE               // Null List interrupt enable
        | BDMARXCON_DIE                // Every Received Frame interrupt enable
        | BDMARXCON_MA_INC             // Memory address increment
        | BDMARXCON_STP_SKP            // Stop/Skip  Frame or Interrupt(=1)
        | 15;                          // Burst size - 1

    // configure Rx MAC
    MACRXCON = MACRXCON_STRIP_CRC;     // Strip CRC enable

    // configure Tx BDMA
    BDMATXCON =
        BDMATXCON_MSL110               // wait to fill 6/8 of the BDMA
        | BDMATXCON_TX_EMPTY           // Buffer Empty interrupt enable
        | BDMATXCON_NOIE               // Not Owner interrupt enable
        | BDMATXCON_NLIE               // Null List interrupt enable
        | BDMATXCON_CPIE               // Control Frame Sent interrupt enable
        | BDMATXCON_STP_SKP            // Stop/Skip Frame or Interrupt
        | 15;                          // Burst size - 1

    // configure Tx MAC
    MACTXCON = MACTXCON_EN_COMP;       // Tx Complete interrupt enable
}


//======================================================================
//  Ethernet controller interrupt support.

#define ALL_ETH_IRQ     0           // enable all the Ethernet interrupts

static cyg_uint32
s3c4510_BDMA_Rx_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_uint32 bdmaStat = BDMASTAT;

#if ALL_ETH_IRQ
    BDMASTAT = bdmaStat & BDMASTAT_RX_ALL;
#else
    BDMASTAT = bdmaStat & (BDMASTAT_RX_ALL | BDMASTAT_TX_ALL);
#endif
    cyg_drv_interrupt_acknowledge(vector);

    STAT_DRV_INC(BDMA_Rx_Cnt, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_RDF), BRxRDF, 1); 
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_NL), BRxNLErr, 1); 
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_NO), BRxNOErr, 1); 
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_MSO), BRxMSOErr, 1); 
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_EMPTY), BRxEmptyErr, 1); 
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_SEARLY), BRxSEarly, 1); 
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_RX_FRF), BRxFRF, 1); 
    STAT_DRV_INC(BRxNRF[(bdmaStat >> 8) & 0xff], 1);

#if !ALL_ETH_IRQ
    // we're collecting Tx stats in the Rx isr.  This can lead to undercounts,
    // but it saves processing the Tx interrupt.
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_CCP), BTxCCP, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_NL), BTxNLErr, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_NO), BTxNOErr, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_EMPTY), BTxEmptyErr, 1);
#endif

    return CYG_ISR_HANDLED | CYG_ISR_CALL_DSR;
}


#ifdef CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
#if ALL_ETH_IRQ
static cyg_uint32
s3c4510_MAC_Rx_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(vector);

    STAT_DRV_INC(MAC_Rx_Cnt, 1);
    STAT_SNMP_INC(interrupts, 1);

    return CYG_ISR_HANDLED;
}


static cyg_uint32
s3c4510_BDMA_Tx_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_uint32 bdmaStat = BDMASTAT;

    BDMASTAT = bdmaStat & BDMASTAT_TX_ALL;
    cyg_drv_interrupt_acknowledge(vector);

    STAT_DRV_INC(BDMA_Tx_Cnt, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_CCP), BTxCCP, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_NL), BTxNLErr, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_NO), BTxNOErr, 1);
    STAT_DRV_TST_INC((bdmaStat & BDMASTAT_TX_EMPTY), BTxEmptyErr, 1);

    return CYG_ISR_HANDLED;
}
#endif // ALL_ETH_IRQ


static cyg_uint32
s3c4510_MAC_Tx_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(vector);

    STAT_DRV_INC(MAC_Tx_Cnt, 1);
    STAT_SNMP_INC(interrupts, 1);

    return CYG_ISR_HANDLED | CYG_ISR_CALL_DSR;
}


static cyg_handle_t bdmaRxIntrHandle;
static cyg_interrupt bdmaRxIntrObject;

static cyg_handle_t macTxIntrHandle;
static cyg_interrupt macTxIntrObject;

#if ALL_ETH_IRQ
static cyg_handle_t bdmaTxIntrHandle;
static cyg_interrupt bdmaTxIntrObject;

static cyg_handle_t macRxIntrHandle;
static cyg_interrupt macRxIntrObject;
#endif


//----------------------------------------------------------------------
//  Initialize the interrupt handlers.

static void
s3c4510_eth_irq_init(struct eth_drv_sc *sc)
{
    debug1_printf("s3c4510_eth_irq_init()\n");

    cyg_drv_interrupt_create(
        CYGNUM_HAL_INTERRUPT_ETH_BDMA_RX,
        0,
        (cyg_addrword_t) sc,
        s3c4510_BDMA_Rx_isr,
        eth_drv_dsr,
        &bdmaRxIntrHandle,
        &bdmaRxIntrObject);
    cyg_drv_interrupt_attach(bdmaRxIntrHandle);
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_ETH_BDMA_RX);

    cyg_drv_interrupt_create(
        CYGNUM_HAL_INTERRUPT_ETH_MAC_TX,
        0,
        (cyg_addrword_t) sc,
        s3c4510_MAC_Tx_isr,
        eth_drv_dsr,
        &macTxIntrHandle,
        &macTxIntrObject);
    cyg_drv_interrupt_attach(macTxIntrHandle);
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_ETH_MAC_TX);

#if ALL_ETH_IRQ
    cyg_drv_interrupt_create(
        CYGNUM_HAL_INTERRUPT_ETH_BDMA_TX,
        0,
        (cyg_addrword_t) sc,
        s3c4510_BDMA_Tx_isr,
        NULL,
        &bdmaTxIntrHandle,
        &bdmaTxIntrObject);
    cyg_drv_interrupt_attach(bdmaTxIntrHandle);
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_ETH_BDMA_TX);

    cyg_drv_interrupt_create(
        CYGNUM_HAL_INTERRUPT_ETH_MAC_RX,
        0,
        (cyg_addrword_t) sc,
        s3c4510_MAC_Rx_isr,
        NULL,
        &macRxIntrHandle,
        &macRxIntrObject);
    cyg_drv_interrupt_attach(macRxIntrHandle);
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_ETH_MAC_RX);
#endif

#ifdef CYGNUM_HAL_INTERRUPT_PHY
    cyg_drv_interrupt_create(
        CYGNUM_HAL_INTERRUPT_PHY,
        0,
        (cyg_addrword_t) sc,
        MAC_Phy_isr,
        MAC_Phy_dsr,
        &macPhyIntrHandle,
        &macPhyIntrObject);
    cyg_drv_interrupt_attach(macPhyIntrHandle);
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_PHY);
#endif
}
#endif // CYGINT_IO_ETH_INT_SUPPORT_REQUIRED


//======================================================================
// Driver code that interfaces to the TCP/IP stack via the common
// Ethernet interface.

//----------------------------------------------------------------------
//  Called by high-level driver to initialize the Ethernet controller.

static bool
s3c4510_eth_init(struct cyg_netdevtab_entry *tab)
{
    struct eth_drv_sc *sc = (struct eth_drv_sc *)tab->device_instance;
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;

    debug1_printf("s3c4510_eth_init()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    // get the Ethernet Station (MAC) address
    if (!s3c4510_eth_get_esa(sc)) {
        return false;
    }

    // initialize the Rx and Tx buffers
    s3c4510_eth_buf_init(sc);

    // initialize the controller
    s3c4510_eth_ctrl_init(sc);

#ifdef CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
    // initialize the interrupts
    s3c4510_eth_irq_init(sc);
#endif

    // initialize the upper level driver
    (sc->funs->eth_drv->init)(sc, spd->esa);

    return true;
}


//----------------------------------------------------------------------
//  Called by high-level driver to enable Rx and Tx.

static void
s3c4510_eth_start(struct eth_drv_sc *sc, unsigned char *enaddr, int flags)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;

    debug1_printf("s3c4510_eth_start()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    if (!spd->active) {
        MACCON &= ~MACON_HALT_REQ;

        // enable Rx BDMA and MAC
        BDMARXCON |= BDMARXCON_EN;
        MACRXCON |= MACRXCON_RX_EN;

        // enable Tx MAC ?
        if (spd->pTxDone->fd.FrameDataPtr & FRM_OWNERSHIP_BDMA) {
            BDMATXCON |= BDMATXCON_EN;
            MACTXCON |= MACTXCON_TX_EN;
        }

        spd->active = true;
        STAT_SNMP_SET(operational, 3);          // 1 = UNKNOWN, 2 = DOWN, 3 = UP

#ifdef CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
        // enable interrupt processing
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_ETH_BDMA_RX);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_ETH_MAC_TX);
#if ALL_ETH_IRQ
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_ETH_BDMA_TX);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_ETH_MAC_RX);
#endif
#ifdef CYGNUM_HAL_INTERRUPT_PHY
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_PHY);
#endif
#endif // CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
    }
}


//----------------------------------------------------------------------
//  Called by high-level driver to disable Rx and Tx.

static void
s3c4510_eth_stop(struct eth_drv_sc *sc)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;

    debug1_printf("s3c4510_eth_stop()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    // halt Tx and Rx after completing any current frames
    MACCON |= MACON_HALT_REQ;
    spd->active = false;
    STAT_SNMP_SET(operational, 2);          // 1 = UNKNOWN, 2 = DOWN, 3 = UP

#ifdef CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
    // disable interrupt processing
    cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_ETH_BDMA_RX);
    cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_ETH_MAC_TX);
#if ALL_ETH_IRQ
    cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_ETH_BDMA_TX);
    cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_ETH_MAC_RX);
#endif
#ifdef CYGNUM_HAL_INTERRUPT_PHY
    cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_PHY);
#endif
#endif // CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
}


//----------------------------------------------------------------------
//  Called by high-level driver to "control" the hardware.

static int
s3c4510_eth_control(struct eth_drv_sc *sc,
    unsigned long cmd, void *data, int len)
{
    debug1_printf("s3c4510_eth_control: cmd=%x\n", cmd);

    switch (cmd) {
#ifdef ETH_DRV_GET_IF_STATS_UD
    case ETH_DRV_GET_IF_STATS_UD: // UD == UPDATE
        // drop through
#endif
#ifdef ETH_DRV_GET_IF_STATS
    case ETH_DRV_GET_IF_STATS:
#endif
#if defined(ETH_DRV_GET_IF_STATS) || defined(ETH_DRV_GET_IF_STATS_UD)
        if (len < sizeof(struct ether_drv_stats))
            return -1;

        *(struct ether_drv_stats *)data = s3c4510_snmp_stats;
        return 0;
#endif

    case ETH_DRV_SET_MAC_ADDRESS: {
        s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;
        cyg_bool active = spd->active;

        CYG_ASSERTC(spd->pSelf == spd);

        if (ETHER_ADDR_LEN != len)
            return -1;

        if (active)
            s3c4510_eth_stop(sc);

        s3c4510_eth_set_cam(0, data);
        memcpy(spd->esa, data, ETHER_ADDR_LEN);

#ifdef CYGFUN_ESA_PLF_SET
        // save the ESA
        CYGFUN_ESA_PLF_SET(spd);
#endif
        if (active)
            s3c4510_eth_start(sc, data, 0);

        return 0;
    }

    case ETH_DRV_SET_MC_LIST: {
        struct eth_drv_mc_list *pmc = (struct eth_drv_mc_list *)data;

        if (len < sizeof(struct eth_drv_mc_list))
            return -1;

        // can the CAM hold all the addresses ?
        if (pmc->len < CAMEN_NUM_ENTRY)
        {
            int i;

            // delete explicit multicast MAC address filtering
            CAMEN = 0x0001;

            // add the multicast MAC addresses
            for (i = 0; i < pmc->len; ++i) {
                s3c4510_eth_set_cam(i + 1, pmc->addrs[i]);
            }
            return 0;
        }
        // fall thru
    }

    case ETH_DRV_SET_MC_ALL:
        // accept any multicast MAC address
        CAMCON |= CAMCON_GROUP_ACC;

        // delete explicit multicast MAC address filtering
        CAMEN = 0x0001;
        return 0;

    default:
        return -1;
    }
}


//----------------------------------------------------------------------
//  Called by high-level driver to determine how many packets can be sent.
//  Currently, all high-level drivers only care if one packet can be sent.

static int
s3c4510_eth_can_send(struct eth_drv_sc *sc)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;

    debug2_printf("s3c4510_eth_can_send()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    if (!spd->active)
        return 0;

    // is the next Tx buffer free ?
    STAT_DRV_TST_INC(spd->pTxWrite->allocated, TxAllocFail, 1);
    return !spd->pTxWrite->allocated;
}


//----------------------------------------------------------------------
//  Called by high-level driver to send a packet.

static void
s3c4510_eth_send(struct eth_drv_sc *sc,
    struct eth_drv_sg *sg_list, int sg_len, int total_len, unsigned long key)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;
    cyg_uint8 *dest = (cyg_uint8 *)spd->pTxWrite->fd.FrameDataPtr;
    cyg_uint32 len = 0;

    debug2_printf("s3c4510_eth_send()\n");
    CYG_ASSERTC(spd->pSelf == spd);
    CYG_ASSERTC(total_len <= MAX_ETH_FRAME_SIZE);
    CYG_ASSERTC((spd->pTxWrite->fd.FrameDataPtr & FRM_OWNERSHIP_BDMA) == 0x0);
    CYG_ASSERTC(!spd->pTxWrite->allocated);
    CYG_ASSERTC(spd->active);

    spd->pTxWrite->key = key;
    spd->pTxWrite->allocated = true;
    STAT_DRV_INC(tx_send_cnt, 1);
    STAT_SNMP_INC(tx_count, 1);

    // copy data from scatter/gather list into BDMA data buffer
    while (sg_len) {
        memcpy(dest, (void *)sg_list->buf, sg_list->len);
        len += sg_list->len;
        dest += sg_list->len;
        ++sg_list;
        --sg_len;
        CYG_ASSERTC(len <= MAX_ETH_FRAME_SIZE);
    }

    cyg_drv_isr_lock();

    // initialize the frame descriptor
    spd->pTxWrite->fd.Reserved =
        TXFDCON_PADDING_MODE
        | TXFDCON_CRC_MODE
        | TXFDCON_MAC_TX_INT_EN
#if (CYG_BYTEORDER == CYG_MSBFIRST)
        | TXFDCON_BIG_ENDIAN
#else
        | TXFDCON_LITTLE_ENDIAN
#endif
        | TXFDCON_SRC_ADDR_INC
        | TXFDCON_WIDGET_ALIGN00;

    spd->pTxWrite->fd.StatusAndFrameLength = len;

    // give the frame to BDMA
    spd->pTxWrite->fd.FrameDataPtr |= FRM_OWNERSHIP_BDMA;

    // start transmit
    BDMATXCON |= BDMATXCON_EN;
    MACTXCON |= MACTXCON_TX_EN;

    cyg_drv_isr_unlock();

    // point to next free Tx frame
    spd->pTxWrite = (s3c4510_eth_buffer_t *)spd->pTxWrite->fd.NextFD;
    CYG_ASSERTC(spd->pTxWrite->pSelf == spd->pTxWrite);
}


//----------------------------------------------------------------------
//  Called by high-level driver to do the work of copying packets to/from
//  the Ethernet controller at thread time.  Not much to do because the
//  BDMA did most of the work.

static void
s3c4510_eth_deliver(struct eth_drv_sc *sc)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;

    CYG_ASSERTC(spd->pSelf == spd);
    STAT_DRV_INC(Deliver_Cnt, 1);

    // process any Tx complete buffers
    while (spd->pTxDone->allocated
    &&  (spd->pTxDone->fd.FrameDataPtr & FRM_OWNERSHIP_BDMA) == 0x0) {
        s3c4510_eth_process_tx_buf(sc, spd->pTxDone);

        // free the buffer
        spd->pTxDone->allocated = false;

        // next Tx frame descriptor
        spd->pTxDone = (s3c4510_eth_buffer_t *)spd->pTxDone->fd.NextFD;
        CYG_ASSERTC(spd->pTxDone->pSelf == spd->pTxDone);
    }

    // process Rx complete buffers
    while ((spd->pRxRead->fd.FrameDataPtr & FRM_OWNERSHIP_BDMA) == 0x0) {
        s3c4510_eth_process_rx_buf(sc, spd->pRxRead);

        // give the buffer back to the BDMA
        spd->pRxRead->fd.StatusAndFrameLength = 0x0;
        spd->pRxRead->fd.FrameDataPtr |= FRM_OWNERSHIP_BDMA;

        // enable Rx, in case Rx BDMA ran out of buffers
        if (spd->active) {
            BDMARXCON |= BDMARXCON_EN;
            MACRXCON |= MACRXCON_RX_EN;
        }

        // next Rx frame descriptor
        spd->pRxRead = (s3c4510_eth_buffer_t *)spd->pRxRead->fd.NextFD;
        CYG_ASSERTC(spd->pRxRead->pSelf == spd->pRxRead);
    }
}


//----------------------------------------------------------------------
//  Called by high-level driver to receive a packet.
//  Currently all high-level drivers call this function from their
//  'sc->funs->eth_drv->recv()' function.

static void
s3c4510_eth_recv(struct eth_drv_sc *sc,
    struct eth_drv_sg *sg_list, int sg_len)
{
    s3c4510_priv_data_t *spd = (s3c4510_priv_data_t *)sc->driver_private;
    cyg_uint8 *source = (cyg_uint8 *)&spd->pRxRead->data[0];
    cyg_uint32 nBytes = 0;

    debug2_printf("s3c4510_eth_recv()\n");
    CYG_ASSERTC(spd->pSelf == spd);

    // copy data from eth buffer into scatter/gather list
    while (sg_len) {
        if (sg_list->buf)
            memcpy((unsigned char *)sg_list->buf, source, sg_list->len);
        source += sg_list->len;
        nBytes += sg_list->len;
        ++sg_list;
        --sg_len;
    }
    CYG_ASSERTC(nBytes <= (spd->pRxRead->fd.StatusAndFrameLength & 0xffff));

    STAT_SNMP_INC(rx_deliver, 1);
}


//----------------------------------------------------------------------
//  Called by high-level driver to poll the Ethernet controller.

static void
s3c4510_eth_poll(struct eth_drv_sc *sc)
{
    STAT_DRV_INC(Poll_Cnt, 1);

    // collect stats
    s3c4510_BDMA_Rx_isr(CYGNUM_HAL_INTERRUPT_ETH_BDMA_RX, (cyg_addrword_t)sc);
    // s3c4510_BDMA_Tx_isr(CYGNUM_HAL_INTERRUPT_ETH_BDMA_TX, (cyg_addrword_t)sc);

    s3c4510_eth_deliver(sc);
}


//----------------------------------------------------------------------
//  Called by high-level driver to return the Rx interrupt vector number.

static int
s3c4510_eth_int_vector(struct eth_drv_sc *sc)
{
    return CYGNUM_HAL_INTERRUPT_ETH_BDMA_RX;
}
