#ifndef CYGONCE_PHY_INL
#define CYGONCE_PHY_INL

// functions to read/write Phy chip registers via MII interface
// on 32c5000.

#define PHYREGWRITE     0x0400
#define MiiStart        0x0800

#define PhyStatus_LinkUp      0x0001
#define PhyStatus_FullDuplex  0x0002
#define PhyStatus_100Mb       0x0004


static void
MiiStationWrite(cyg_uint32 RegAddr, cyg_uint32 PhyAddr, cyg_uint32 PhyWrData)
{
    STADATA = PhyWrData;
    STACON = RegAddr | (PhyAddr << 5) | MiiStart | PHYREGWRITE;
    while (STACON & MiiStart)
        continue;

    //debug1_printf("PHY Wr %x:%02x := %04x\n",PhyAddr, RegAddr, PhyWrData) ;
}

static cyg_uint32
MiiStationRead(cyg_uint32 RegAddr, cyg_uint32 PhyAddr)
{
    cyg_uint32 PhyRdData;

    STACON = RegAddr | (PhyAddr << 5) | MiiStart;
    while (STACON & MiiStart)
        continue;

    PhyRdData = STADATA;
    //debug1_printf("PHY Rd %x:%02x  %04x\n",PhyAddr,RegAddr,PhyRdData) ;
    return PhyRdData;
}

#ifdef CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
#ifdef CYGNUM_HAL_INTERRUPT_PHY
static cyg_handle_t macPhyIntrHandle;
static cyg_interrupt macPhyIntrObject;
static cyg_uint32 s3c4510_MAC_Phy_Isr_Cnt;
static unsigned s3c4510_linkStatus;

static void PhyInterruptAck(void);

static cyg_uint32
MAC_Phy_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(vector);
    PhyInterruptAck();
    ++s3c4510_MAC_Phy_Isr_Cnt;

    return CYG_ISR_HANDLED | CYG_ISR_CALL_DSR;
}

static void
MAC_Phy_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    struct eth_drv_sc *sc = (struct eth_drv_sc *sc)data;

    s3c4510_linkStatus = PhyStatus();
    if (s3c4510_linkStatus & PhyStatus_FullDuplex)
        MACCON |= MACON_FULL_DUP;
    else
        MACCON &= ~MACON_FULL_DUP;

    // 1 = UNKNOWN, 2 = SIMPLEX, 3 = DUPLEX
    SET_STAT(duplex, (s3c4510_linkStatus & PhyStatus_FullDuplex) ? 3 : 2);

    // 1 = UNKNOWN, 2 = DOWN, 3 = UP
    SET_STAT(operational, (s3c4510_linkStatus & PhyStatus_LinkUp) ? 3 : 2);

    SET_STAT(speed, (s3c4510_linkStatus & PhyStatus_100Mb) ? 100*1000*1000 : 10*1000*1000);
}

#endif // CYGNUM_HAL_INTERRUPT_PHY
#endif // CYGINT_IO_ETH_INT_SUPPORT_REQUIRED
#endif // CYGONCE_PHY_INL
