//==========================================================================
//
//      devs_eth_arm_snds100.inl
//
//      SNDS100 platform dependent support for the Samsung S3C4510x
//      Ethernet Controller
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):   jskov
// Contributors:jskov, mchecky
// Date:        2003-05-05
//
//####DESCRIPTIONEND####
//==========================================================================

#ifdef CYGPKG_DEVS_ETH_ARM_SNDS100_ETH0
#include <cyg/hal/hal_if.h>
#include <errno.h>


// -------------------------------------------------------------------------
// Platform specific Ethernet station address access

#ifdef CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_EEPROM
# include <cyg/io/eeprom.h>

# define CYGFUN_ESA_PLF_GET(spd) snds100_esa_plf_get(spd)
# define ESA_EEPROM_ADDR    (0)

static cyg_bool
snds100_esa_plf_get(s3c4510_priv_data_t* spd)
{
    if (EEPROM_Init() != ENOERR)
    {
        return false;
    }

    return (EEPROM_Read(ESA_EEPROM_ADDR, spd->esa, sizeof(spd->esa)) == ENOERR);
}

# ifdef CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_EEPROM_WRITE
#  define CYGFUN_ESA_PLF_SET(spd) snds100_esa_plf_set(spd)

static cyg_bool
snds100_esa_plf_set(s3c4510_priv_data_t* spd)
{
    return (EEPROM_Write(ESA_EEPROM_ADDR, spd->esa, sizeof(spd->esa)) == ENOERR);
}

# endif // CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_EEPROM_WRITE
#endif // CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_EEPROM


// -------------------------------------------------------------------------
// Ethernet station address from Redboot Flash configuration

#ifdef CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_REDBOOT
# ifdef CYGPKG_REDBOOT
#  include <pkgconf/redboot.h>

#  ifdef CYGSEM_REDBOOT_FLASH_CONFIG
#   include <redboot.h>
#   include <flash_config.h>

RedBoot_config_option(
        "Set " CYGDAT_DEVS_ETH_ARM_SNDS100_ETH0_NAME " network hardware address [MAC]",
        eth0_esa,
        ALWAYS_ENABLED, true,
        CONFIG_BOOL, false);

RedBoot_config_option(
        CYGDAT_DEVS_ETH_ARM_SNDS100_ETH0_NAME " network hardware address [MAC]",
        eth0_esa_data,
        "eth0_esa", true,
        CONFIG_ESA, 0);
#  endif // CYGSEM_REDBOOT_FLASH_CONFIG
# endif // CYGPKG_REDBOOT


# ifdef CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT
// Note that this section *is* active in an application, outside RedBoot,
// where the above section is not included.

#  define CYGFUN_ESA_REDBOOT_GET(spd) snds100_esa_redboot_get(spd)

// check the Redboot Flash configuration for the ESA
static cyg_bool
snds100_esa_redboot_get(s3c4510_priv_data_t* spd)
{
    cyg_bool set_esa;
    int ok;

    ok = CYGACC_CALL_IF_FLASH_CFG_OP(CYGNUM_CALL_IF_FLASH_CFG_GET,
                                     "eth0_esa",
                                     &set_esa,
                                     CYGNUM_FLASH_CFG_OP_CONFIG_BOOL);
    if (ok && set_esa) {
        ok = CYGACC_CALL_IF_FLASH_CFG_OP(CYGNUM_CALL_IF_FLASH_CFG_GET,
                                         "eth0_esa_data",
                                         spd->esa,
                                         CYGNUM_FLASH_CFG_OP_CONFIG_ESA);
    }

    return ok && set_esa;
}

# endif // CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT
#endif // CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_REDBOOT


// device private data
static s3c4510_priv_data_t s3c4510_eth0_priv_data = { 
    .pSelf = &s3c4510_eth0_priv_data, 
#ifdef CYGFUN_DEVS_ETH_ARM_SNDS100_ETH0_ESA_SET
    .esa = { CYGDAT_DEVS_ETH_ARM_SNDS100_ETH0_ESA },
    .hardwired_esa = true,
#endif
};


// standard Ethernet driver interface
ETH_DRV_SC(s3c4510_eth0_sc,
           &s3c4510_eth0_priv_data,
           CYGDAT_DEVS_ETH_ARM_SNDS100_ETH0_NAME,
           s3c4510_eth_start,
           s3c4510_eth_stop,
           s3c4510_eth_control,
           s3c4510_eth_can_send,
           s3c4510_eth_send,
           s3c4510_eth_recv,
           s3c4510_eth_deliver,
           s3c4510_eth_poll,
           s3c4510_eth_int_vector);

NETDEVTAB_ENTRY(s3c4510_eth0_netdev, 
                "s3c4510_" CYGDAT_DEVS_ETH_ARM_SNDS100_ETH0_NAME,
                s3c4510_eth_init, 
                &s3c4510_eth0_sc);

#endif // CYGPKG_DEVS_ETH_ARM_SNDS100_ETH0
