//-----------------------------------------------------------------------------
//
//  FILE: i2c.h
//
//  DESCRIPTION:
//      This header file contains the typedefs, #defines, and function
//      prototypes for the eCos device driver routines for the Samsung
//      S3C4510x I2C bus controller.
//
//  MODIFICATION HISTORY:
//
//

#ifndef CYGONCE_I2C_H
#define CYGONCE_I2C_H

#include <pkgconf/system.h>
#include <cyg/infra/cyg_type.h>
#include <errno.h>


#define I2C_FREQ(freq)      ((CYGNUM_HAL_CPUCLOCK/(freq) - 3)/16)
#define I2C_RD              (0x01)
#define I2C_WR              (0x00)

#define I2C_STATUS_SUCCESS      (ENOERR)
#define I2C_STATUS_MUTEX        (-EINTR)
#define I2C_STATUS_BUSY         (-EBUSY)
#define I2C_STATUS_ADDR_NAK     (-1000)
#define I2C_STATUS_DATA_NAK     (-1001)

typedef struct I2C_Msg_s
{
    cyg_uint8   devAddr;
    cyg_int8    status;
    cyg_uint8*  pBuf;
    cyg_uint32  bufSize;
} I2C_Msg_t;

// function prototypes
__externC int
I2C_Init(cyg_uint32 prescale);
__externC int
I2C_Transfer(size_t nMsg, I2C_Msg_t* pMsgs);

#endif
